//
// Path Tracer by SSJX (https://ssjx.co.uk) 
//
// Based on the Postcard Sized Path Tracer by Andrew Kensler.
//
// Websites that were useful in the making of this program:
//
// https://fabiensanglard.net/postcard_pathtracer/index.html
//
// https://mattwarren.org/2019/03/01/Is-CSharp-a-low-level-language/
//
// https://medium.com/@cfsamson/from-48s-to-5s-optimizing-a-350-line-pathtracer-in-rust-191ab4a1a412
//
// Thank you!
//
// Note some of the comments are from the C++ version, mostly where they apply to the maths bits...
//
// Compiles with:
// 	dmd -i pixar.d
//
// Version History
// ===============
// 0.1a (15/02/2024)
// + Recompiled with LDC, even faster now!
//
// 0.1 (14/02/2024)
// + First release

import std.stdio;
import std.random;
import std.math;
import std.algorithm;	// min
import std.parallelism;
import std.datetime.stopwatch;

import bmp;

struct Vec {
    float x, y, z;

    this(float v) { x = y = z = v; }

    this(float a, float b) {
      x = a;
      y = b;
      z = 0;
    }

   this(float a, float b, float c) {
      x = a;
      y = b;
      z = c;
    }

	Vec opBinary(string s)(Vec r) if (s == "+"){
		return Vec(x + r.x, y + r.y, z + r.z);
    }

	Vec opBinary(string s)(Vec r) if (s == "*"){
		return Vec(x * r.x, y * r.y, z * r.z); 
    }

	Vec opBinary(string s)(float r) if (s == "*"){
		return Vec(x * r, y * r, z * r);
    }
	
	Vec opBinary(string s)(float r) if (s == "+"){
		return Vec(x + r, y + r, z + r);
    }
	
	float opBinary(string s)(Vec r) if (s == "%"){
		return (x * r.x) + (y * r.y) + (z * r.z);
    }
};

// Inverse Square Root
Vec inverse(Vec r){
	float sq=1 / sqrt( r%r);
	return r*sq;
}


//
float randomVal() { return uniform(0.0f, 1.0f); }

// Rectangle CSG equation. Returns minimum signed distance from
// space carved by
// lowerLeft vertex and opposite rectangle vertex upperRight.
float BoxTest(Vec position, Vec lowerLeft, Vec upperRight) {
  lowerLeft = position + (lowerLeft * -1.0f);
  upperRight = upperRight + (position * -1.0f);
  return -min(
			min(min(lowerLeft.x, upperRight.x), min(lowerLeft.y, upperRight.y)),
			min(lowerLeft.z, upperRight.z)
		  );
}

const int HIT_NONE=0;
const int HIT_LETTER=1;
const int HIT_WALL=2;
const int HIT_SUN=3;

static char[] letters =               	// 15 two points lines
		"5O5_"~"5W9W"~"5_9_"~         	// P (without curve)
		"AOEO"~"COC_"~"A_E_"~         	// I
		"IOQ_"~"I_QO"~                	// X
		"UOY_"~"Y_]O"~"WW[W"~         	// A
		"aOa_"~"aWeW"~"a_e_"~"cWiO" 	// R (without curve)
		.dup;	// String to char array

static Vec[2] curves = [Vec(-11, 6), Vec(11, 6)];

// Sample the world using Signed Distance Fields.
float QueryDatabase(Vec position, ref int hitType) {
	float distance=float.max; // 1e9;
	Vec f = position; // Flattened position (z=0)
	f.z = 0;
	
	// Letters were here!	
		
	for (int i = 0; i < letters.length; i += 4) {
		Vec begin = Vec(letters[i] - 79, letters[i + 1] - 79 ) * 0.5;

		Vec e = Vec(letters[i + 2] - 79, letters[i + 3] - 79) *  0.5 + begin * -1;

		Vec o = f + (begin + e * min(-min((begin + f * -1 ) % e / (e % e), 0), 1) ) * -1;

		distance = min(distance, o % o); // compare squared distance.
	}
	distance = sqrt(distance); // Get real distance, not square distance.

	// Two curves (for P and R in PixaR) with hard-coded locations.
	// Vec[2] curves = [Vec(-11, 6), Vec(11, 6)];
	for (int i = 2; i--;) {
		Vec o = f + curves[i] * -1;

		float temp;
		if (o.x > 0) {
			temp = fabs(sqrt(o % o) - 2);
		} else {
			o.y += o.y > 0 ? -2 : 2;
			temp = sqrt(o % o);
		}
		distance = min(distance, temp);
	}
	
	distance = pow(pow(distance, 8) + pow(position.z, 8), .125) - .5;
	hitType = HIT_LETTER;

	float roomDist ;
	roomDist = min(	// min(A,B) = Union with Constructive solid geometry
			//-min carves an empty space
			-min(// Lower room
				BoxTest(position, Vec(-30, -0.5, -30), Vec(30, 18, 30)),
				// Upper room
				BoxTest(position, Vec(-25, 17, -25), Vec(25, 20, 25))
			),
			BoxTest( // Ceiling "planks" spaced 8 units apart.
				Vec(fmod(fabs(position.x), 8), position.y, position.z),
				Vec(1.5, 18.5, -25),
				Vec(6.5, 20, 25)
			)
	);
	if (roomDist < distance){
		distance = roomDist; 
		hitType = HIT_WALL;
	}

	float sun = 19.9 - position.y ; // Everything above 19.9 is light source.
	
	if (sun < distance){
		distance = sun;
		hitType = HIT_SUN;
	}
	
	return distance;
}

// Perform signed sphere marching
// Returns hitType 0, 1, 2, or 3 and update hit position/normal
int RayMarching(Vec origin, Vec direction, ref Vec hitPos, ref Vec hitNorm) {
	int hitType = HIT_NONE;
	int noHitCount = 0;
	float d; // distance from closest object in world.

	// Signed distance marching
	for (float total_d=0; total_d < 100; total_d += d){
		if ((d = QueryDatabase(hitPos = origin + direction * total_d, hitType)) < .01 || ++noHitCount > 99){
			hitNorm =inverse(Vec(QueryDatabase(hitPos + Vec(.01, 0), noHitCount) - d,
			QueryDatabase(hitPos + Vec(0, .01), noHitCount) - d,
			QueryDatabase(hitPos + Vec(0, 0, .01), noHitCount) - d));
			return hitType; // Weird return statement where a variable is also updated.
		}
	}
	return 0;
}

Vec Trace(Vec origin, Vec direction) {
	Vec sampledPosition=0; 
	Vec normal=0;
	Vec color=0;
	Vec attenuation = 1;
	
	// Vec lightDirection(!Vec(.6, .6, 1)); // Directional light
	Vec lightDirection=inverse(Vec(0.6, 0.6, 1)); // Directional light

	foreach(bounceCount;0..4) {
		int hitType = RayMarching(origin, direction, sampledPosition, normal);

		if (hitType == HIT_NONE){
			break; // No hit. This is over, return color.
		}

		if (hitType == HIT_LETTER) { // Specular bounce on a letter. No color acc.
			direction = direction + normal * ( normal % direction * -2);
			origin = sampledPosition + direction * 0.1;
			attenuation = attenuation * 0.2; // Attenuation via distance traveled.
		}

		if (hitType == HIT_WALL) { // Wall hit uses color yellow?
			float incidence = normal % lightDirection;
			float p = 6.283185 * randomVal();
			float c = randomVal();
			float s = sqrt(1 - c);
			float g = normal.z < 0 ? -1 : 1;
			float u = -1 / (g + normal.z);
			float v = normal.x * normal.y * u;
			direction = Vec(v, g + normal.y * normal.y * u, -normal.y) * (cos(p) * s)
			+Vec(1 + g * normal.x * normal.x * u, g * v, -g * normal.x) * (sin(p) * s) + normal * sqrt(c);

			origin = sampledPosition + direction * 0.1f;

			attenuation = attenuation * 0.2f;

			if (incidence > 0 && RayMarching(sampledPosition + normal * 0.1f, lightDirection, sampledPosition, normal) == HIT_SUN){
				color = color + attenuation * Vec(500, 400, 100) * incidence;
			}
		}

		if (hitType == HIT_SUN) { //
			color = color + attenuation * Vec(50, 80, 100); 
			break; // Sun Color
		}
	}
	return color;
}

void main() {
	// Change these values to make a larger more detailed image.
	// Command line options maybe in the next version!
	const int w = 640;
	const int h = 480;
	const int samplesCount = 8;

	Vec position=Vec(-22, 5, 25);
	Vec goal = inverse( Vec(-3, 4, 0) + position * -1 );
	Vec left = inverse( Vec(goal.z, 0, -goal.x)) * (1. / w) ;

	// Cross-product to get the up vector
	Vec up=Vec(goal.y * left.z - goal.z * left.y,
	goal.z * left.x - goal.x * left.z,
	goal.x * left.y - goal.y * left.x);

	ubyte[] output=new ubyte[w*h*3];
	
	writeln("Path Tracer v0.1a (15/02/2024) by ssjx (https://ssjx.co.uk)");
	writeln("===========================================================");
	writeln("Built with the ",__VENDOR__," Compiler");
	writeln("Please wait, this may take some time...");
	
	StopWatch sw;
	sw.start();
	
	// ==== PARALLEL ====
	struct mypoint{
		int x;
		int y;
		int pos;
	}
	mypoint[] points=new mypoint[w*h];
	
	auto c=0;
	foreach_reverse(y;0..h){
		auto px=y*(w*3);
		foreach_reverse(x;0..w){
			points[c].x=x;
			points[c].y=y;
			points[c].pos=px;
			px+=3;
			c++;
		}
	}
	
	foreach(pt;parallel(points)){
		auto x=pt.x;
		auto y=pt.y;
		auto px=pt.pos;
		
		Vec color=0;
		foreach(p;0..samplesCount){
			color = color + Trace(position, inverse(goal + left * (x - w / 2 + randomVal()) + up * (y - h / 2 + randomVal())));
		}
		// Reinhard tone mapping
		color = color * (1.0f / samplesCount) + 14.0f / 241;
		Vec o = color + 1;
		color = Vec(color.x / o.x, color.y / o.y, color.z / o.z) * 255;
	
		// Write our bitmap
		output[px]=cast(ubyte)color.z;
		output[px+1]=cast(ubyte)color.y;
		output[px+2]=cast(ubyte)color.x;
	}
	
	// ==== SINGLE ====	
	// foreach_reverse(y;0..h){
		// auto px=y*(w*3);
		// foreach_reverse(x;0..w){
			// Vec color=0;
			// foreach(p;0..samplesCount){
				// color = color + Trace(position, inverse(goal + left * (x - w / 2 + randomVal()) + up * (y - h / 2 + randomVal())));
			// }
			// // Reinhard tone mapping
			// color = color * (1.0f / samplesCount) + 14.0f / 241;
			// Vec o = color + 1;
			// color = Vec(color.x / o.x, color.y / o.y, color.z / o.z) * 255;
		
			// // Write our bitmap
			// output[px]=cast(ubyte)color.z;
			// output[px+1]=cast(ubyte)color.y;
			// output[px+2]=cast(ubyte)color.x;
			// px+=3;
		// }
	// }
	// ========	
	long msecs = sw.peek.total!"msecs";
	writeln("Elapsed time: ",msecs,"ms");
	
	save_bmp(w,h,output,"output");
}

